#ifndef SEMANTICO_H
#define SEMANTICO_H

#include "error.h"
#include "lexer.h"
#include "parser.h"
#include "semantico_aux.h"
#include <set>
#include <map>
#include <string_view>
#include <vector>

struct datos_funcion;

struct analisis_funcion {
   tipo_evaluado retorno;
   std::map<const expresion*, tipo_evaluado> tipo_resultante;
   std::map<const expresion_termino*, const sentencia_declaracion*> variable_referida;
   std::map<const expresion_llamada*, const datos_funcion*> funcion_referida;
   std::map<const expresion_termino*, int> entero_referido;
   std::map<const expresion_termino*, std::string> cadena_referida;
   std::map<const sentencia_declaracion*, tipo_evaluado> tipo_variable;
   std::map<const sentencia*, const sentencia*> ciclo_afectado;
};

struct datos_funcion {
   const declaracion_funcion& decl;
   analisis_funcion analisis;
};

struct tabla_simbolos {
   std::map<std::string_view, datos_funcion> funciones;
};

struct pila_simbolos {
   const tabla_simbolos& tabla;
   std::vector<const sentencia*> ciclos;
   std::vector<std::map<std::string_view, const sentencia_declaracion*>> bloques{1};

   const datos_funcion* busca_funcion(const std::string_view& s) const {
      auto iter = tabla.funciones.find(s);
      return (iter != tabla.funciones.end( ) ? &iter->second : nullptr);
   }

   const sentencia_declaracion* busca_variable(const std::string_view& s) const {
      for (int i = bloques.size( ) - 1; i >= 0; --i) {
         if (auto iter = bloques[i].find(s); iter != bloques[i].end( )) {
            return iter->second;
         }
      }
      return nullptr;
   }
};

tipo_evaluado evalua(const expresion_termino* ex, auto& analisis, auto& pila) {
   if (ex->token.tipo == IDENTIFICADOR) {
      if (auto decl = pila.busca_variable(ex->token.vista); decl != nullptr) {
         analisis.variable_referida[ex] = decl;
         return analisis.tipo_resultante[ex] = analisis.tipo_variable[decl];
      }
      throw error("El identificador no denota una variable", ex->vista);
   } else if (ex->token.tipo == LITERAL_ENTERA) {
      analisis.entero_referido[ex] = evalua_entero(ex->token.vista);
      return analisis.tipo_resultante[ex] = { INT, VALOR, TEMPORAL };
   } else if (ex->token.tipo == LITERAL_CADENA) {
      analisis.cadena_referida[ex] = evalua_cadena(ex->token.vista);
      if (analisis.cadena_referida[ex].size( ) == 1) {
         return analisis.tipo_resultante[ex] = { CHAR, VALOR, TEMPORAL };
      } else if (analisis.cadena_referida[ex].size( ) > 1) {
         return analisis.tipo_resultante[ex] = { CHAR, ARREGLO, TEMPORAL, int(analisis.cadena_referida[ex].size( )) };
      } else {
         throw error("Literal de cadena no corresponde con un tipo de arreglo soportado", ex->vista);
      }
   } else {
      throw error("Termino inesperado en expresion", ex->vista);
   }
}

tipo_evaluado evalua(const expresion_binaria* ex, auto& analisis, auto& pila) {
   // por hacer (tarea 6)
}

tipo_evaluado evalua(const expresion_prefija* ex, auto& analisis, auto& pila) {
   // por hacer (tarea 6)
}

tipo_evaluado evalua(const expresion_posfija* ex, auto& analisis, auto& pila) {
   // por hacer (tarea 6)
}

tipo_evaluado evalua(const expresion_llamada* ex, auto& analisis, auto& pila) {
   if (es_tipo(ex->funcion.tipo)) {
      if (ex->parametros.size( ) == 1) {
         auto temp = evalua(ex->parametros[0], analisis, pila);
         if ((ex->funcion.tipo == INT || ex->funcion.tipo == CHAR) && (temp == tipo_evaluado(INT, VALOR) || temp == tipo_evaluado(CHAR, VALOR))) {
            return analisis.tipo_resultante[ex] = { ex->funcion.tipo, VALOR, TEMPORAL };
         }
      }
      throw error("Moldeado invalido", ex->vista);
   } else if (ex->funcion.tipo == PRINT || ex->funcion.tipo == SCAN) {
      if (ex->parametros.size( ) == 1) {
         auto temp = evalua(ex->parametros[0], analisis, pila);
         if (ex->funcion.tipo == PRINT && (temp == tipo_evaluado(INT, VALOR) || temp == tipo_evaluado(CHAR, VALOR) || temp == tipo_evaluado(CHAR, ARREGLO))) {
            return analisis.tipo_resultante[ex] = { VOID, VALOR, TEMPORAL };
         } else if (ex->funcion.tipo == SCAN && temp.vida == VARIABLE && (temp == tipo_evaluado(INT, VALOR) || temp == tipo_evaluado(CHAR, VALOR))) {
            return analisis.tipo_resultante[ex] = { VOID, VALOR, TEMPORAL };
         }
      }
      throw error("Llamada invalida a funcion incorporada", ex->vista);
   } else {
      if (pila.busca_variable(ex->funcion.vista) != nullptr) {
         throw error("Variable usada como funcion", ex->vista);
      }
      auto info = analisis.funcion_referida[ex] = pila.busca_funcion(ex->funcion.vista);
      if (info == nullptr) {
         throw error("Funcion no declarada", ex->vista);
      } else if (info->decl.parametros.size( ) != ex->parametros.size( )) {
         throw error("Llamada con numero incorrecto de parametros", ex->vista);
      }
      for (int i = 0; i < ex->parametros.size( ); ++i) {
         if (evalua(ex->parametros[i], analisis, pila) != info->analisis.tipo_variable.find(info->decl.parametros[i].get( ))->second) {
            throw error("Tipo de argumento incompatible con parametro", ex->vista);
         }
      }
      return analisis.tipo_resultante[ex] = info->analisis.retorno;
   }
}

tipo_evaluado evalua(const expresion_indexacion* ex, auto& analisis, auto& pila) {
   auto izq = evalua(ex->ex, analisis, pila), dentro = evalua(ex->indice, analisis, pila);
   if (izq.modificador == VALOR) {
      throw error("Solo se permite indexar arreglos y apuntadores", ex->vista);
   } else if (dentro != tipo_evaluado(INT, VALOR)) {
      throw error("Solo se pueden usar enteros como indices", ex->vista);
   }
   return analisis.tipo_resultante[ex] = tipo_evaluado(izq.tipo_base, VALOR, (izq.modificador == APUNTADOR ? VARIABLE : izq.vida));
}

tipo_evaluado evalua(const expresion_secuencia* ex, auto& analisis, auto& pila) {
   if (ex->valores.empty( )) {
      throw error("No se permiten literales de secuencia vacias", ex->vista);
   }
   auto primero = evalua(ex->valores[0], analisis, pila);
   if (primero != tipo_evaluado(INT, VALOR) && primero != tipo_evaluado(CHAR, VALOR)) {
      throw error("Solo se permiten arreglos de enteros y caracteres", ex->vista);
   }
   for (int i = 1; i < ex->valores.size( ); ++i) {
      if (primero != evalua(ex->valores[i], analisis, pila)) {
         throw error("Todos los elementos de la secuencia deben ser del mismo tipo", ex->vista);
      }
   }
   return analisis.tipo_resultante[ex] = tipo_evaluado(primero.tipo_base, ARREGLO, TEMPORAL, ex->valores.size( ));
}

void evalua(const sentencia_expresion* s, auto& analisis, auto& pila) {
   evalua(s->ex, analisis, pila);
}

void evalua(const sentencia_declaracion* s, auto& analisis, auto& pila) {
   analisis.tipo_variable[s] = evalua_como_tipo(s->tipo, LOCAL);
   if (!pila.bloques.back( ).emplace(s->nombre.vista, s).second) {
      throw error("Redeclaracion de variable", s->vista);
   }
   if (s->inicializador != nullptr && evalua(s->inicializador, analisis, pila) != analisis.tipo_variable[s]) {
      throw error("Inicializador de tipo incorrecto", s->vista);
   }
}

void evalua(const sentencia_if* s, auto& analisis, auto& pila) {
   if (evalua(s->condicion, analisis, pila) != tipo_evaluado(INT)) {
      throw error("Tipo incorrecto en condicion", s->vista);
   }
   for (auto parte : { &s->parte_si, &s->parte_no }) {
      pila.bloques.emplace_back( );
      for (const auto& actual : *parte) {
         evalua(actual, analisis, pila);
      }
      pila.bloques.pop_back( );
   }
}

void evalua(const sentencia_while* s, auto& analisis, auto& pila) {
   if (evalua(s->condicion, analisis, pila) != tipo_evaluado(INT)) {
      throw error("Tipo incorrecto en condicion", s->vista);
   }
   pila.bloques.emplace_back( ), pila.ciclos.emplace_back(s);
   for (const auto& actual : s->cuerpo) {
      evalua(actual, analisis, pila);
   }
   pila.bloques.pop_back( ), pila.ciclos.pop_back( );
}

void evalua(const sentencia_do* s, auto& analisis, auto& pila) {
   pila.bloques.emplace_back( ), pila.ciclos.emplace_back(s);
   for (const auto& actual : s->cuerpo) {
      evalua(actual, analisis, pila);
   }
   pila.bloques.pop_back( ), pila.ciclos.pop_back( );
   if (evalua(s->condicion, analisis, pila) != tipo_evaluado(INT)) {
      throw error("Tipo incorrecto en condicion", s->vista);
   }
}

void evalua(const sentencia_for* s, auto& analisis, auto& pila) {
   pila.bloques.emplace_back( ), pila.ciclos.emplace_back(s);
   if (s->inicializacion != nullptr) {
      evalua(s->inicializacion.get( ), analisis, pila);
   }
   if (s->condicion != nullptr && evalua(s->condicion, analisis, pila) != tipo_evaluado(INT)) {
      throw error("Tipo incorrecto en condicion", s->vista);
   }
   if (s->actualizacion != nullptr) {
      evalua(s->actualizacion, analisis, pila);
   }
   for (const auto& actual : s->cuerpo) {
      evalua(actual, analisis, pila);
   }
   pila.bloques.pop_back( ), pila.ciclos.pop_back( );
}

void evalua(const sentencia_return* s, auto& analisis, auto& pila) {
   if (s->ex != nullptr) {
      if (evalua(s->ex, analisis, pila) != analisis.retorno) {
         throw error("Sentencia return con tipo incompatible", s->vista);
      }
   } else if (analisis.retorno != tipo_evaluado(VOID)) {
      throw error("Sentencia return sin valor en funcion no void", s->vista);
   }
}

void evalua(const sentencia_break* s, auto& analisis, auto& pila) {
   if (pila.ciclos.empty( )) {
      throw error("Sentencia break fuera de ciclo", s->vista);
   }
   analisis.ciclo_afectado[s] = pila.ciclos.back( );
}

void evalua(const sentencia_continue* s, auto& analisis, auto& pila) {
   if (pila.ciclos.empty( )) {
      throw error("Sentencia continue fuera de ciclo", s->vista);
   }
   analisis.ciclo_afectado[s] = pila.ciclos.back( );
}

void evalua(const sentencia_exit* s, auto& analisis, auto& pila) {
   return;
}

tabla_simbolos semantico(const arbol_sintactico& arbol, const token_anotado& fin_archivo) {
   tabla_simbolos tabla;
   for (const auto& funcion : arbol.funciones) {
      analisis_funcion analisis;
      analisis.retorno = evalua_como_tipo(funcion.retorno, RETORNO);
      std::set<std::string_view> nombres_parametros;
      for (const auto& parametro : funcion.parametros) {
         analisis.tipo_variable.emplace(parametro.get( ), evalua_como_tipo(parametro->tipo, PARAMETRO));
         if (!nombres_parametros.insert(parametro->nombre.vista).second) {
            throw error("Nombre de parametro repetido", parametro->vista);
         }
      }
      if (!tabla.funciones.emplace(funcion.nombre.vista, datos_funcion(funcion, std::move(analisis))).second) {
         throw error("Nombre de funcion repetida", funcion.nombre.vista);
      }
   }
   for (const auto& funcion : arbol.funciones) {
      auto& datos = tabla.funciones.find(funcion.nombre.vista)->second;
      pila_simbolos pila(tabla);
      for (const auto& parametro : datos.decl.parametros) {
         pila.bloques[0][parametro->nombre.vista] = parametro.get( );
      }
      for (const auto& sentencia : datos.decl.cuerpo) {
         evalua(sentencia, datos.analisis, pila);
      }
   }
   if (auto iter = tabla.funciones.find("main"); iter == tabla.funciones.end( ) || iter->second.analisis.retorno != tipo_evaluado(VOID, VALOR) || !iter->second.decl.parametros.empty( )) {
      throw error("Funcion main no declarada o de tipo incorrecto", fin_archivo.vista);
   }
   return tabla;
}

#endif
