#ifndef SEMANTICO_AUX_H
#define SEMANTICO_AUX_H

#include "error.h"
#include "lexer.h"
#include "parser.h"
#include <charconv>
#include <vector>
#include <stddef.h>

enum contexto_tipo {
   RETORNO, PARAMETRO, LOCAL
};

enum modificador_tipo {
   VALOR, APUNTADOR, ARREGLO
};

enum tiempo_vida {
   VARIABLE, TEMPORAL
};

struct tipo_evaluado {
   token tipo_base;
   modificador_tipo modificador = VALOR;
   tiempo_vida vida = TEMPORAL;
   int tamanyo = -1;

   bool operator==(const tipo_evaluado& t) const {
      if (tipo_base != t.tipo_base || modificador != t.modificador) {
         return false;
      } else if (modificador == ARREGLO && (tamanyo != -1 && t.tamanyo != -1 && tamanyo != t.tamanyo)) {
         return false;
      }
      return true;
   }
};

int evalua_entero(const std::string_view& s) {
   if (int res; std::from_chars(s.begin( ), s.end( ), res).ec == std::errc( )) {
      return res;
   }
   throw error("Entero no representable", s);
}

std::string evalua_cadena(const std::string_view& s) {
   std::string res;
   for (int i = 1; i < s.size( ) - 1; ++i) {
      res.push_back((s[i] != '\\' ? s[i] : (s[++i] == '\\' ? '\\' : (s[i] == 'n' ? '\n' : '"'))));
   }
   return res;
}

tipo_evaluado evalua_como_tipo(const std::unique_ptr<expresion>& ex, contexto_tipo contexto) {
   auto vida = (contexto == RETORNO ? TEMPORAL : VARIABLE);
   if (auto b = dynamic_cast<const expresion_termino*>(ex.get( )); b != nullptr) {
      if (es_tipo(b->token.tipo) && (b->token.tipo != VOID || contexto == RETORNO)) {
         return { b->token.tipo, VALOR, vida };
      }
   } else if (auto p = dynamic_cast<const expresion_posfija*>(ex.get( )); p != nullptr) {
      if (auto b = dynamic_cast<const expresion_termino*>(p->ex.get( )); b != nullptr) {
         if (es_tipo(b->token.tipo) && b->token.tipo != VOID && p->op.tipo == DESREFERENCIA) {
            return { b->token.tipo, APUNTADOR, vida };
         }
      }
   } else if (auto p = dynamic_cast<const expresion_indexacion*>(ex.get( )); p != nullptr) {
      if (auto b = dynamic_cast<const expresion_termino*>(p->ex.get( )); b != nullptr) {
         if (auto x = dynamic_cast<const expresion_termino*>(p->indice.get( )); x != nullptr) {
            if (es_tipo(b->token.tipo) && b->token.tipo != VOID && x->token.tipo == LITERAL_ENTERA && contexto == LOCAL) {
               if (int tam = evalua_entero(x->vista); tam >= 2) {
                  return { b->token.tipo, ARREGLO, vida, tam };
               }
            }
         }
      }
   }
   throw error("Tipo incorrecto", ex->vista);
}

tipo_evaluado evalua(const std::unique_ptr<expresion>& ex, auto&... params) {
   if (auto p = dynamic_cast<const expresion_termino*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_binaria*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_prefija*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_posfija*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_llamada*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_indexacion*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const expresion_secuencia*>(ex.get( )); p != nullptr) {
      return evalua(p, params...);
   } else {
      return { };
   }
}

void evalua(const std::unique_ptr<sentencia>& s, auto&... params) {
   if (auto p = dynamic_cast<const sentencia_expresion*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_declaracion*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_if*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_while*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_do*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_for*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_return*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_break*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_continue*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   } else if (auto p = dynamic_cast<const sentencia_exit*>(s.get( )); p != nullptr) {
      return evalua(p, params...);
   }
}

#endif
